"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSData = void 0;
const commons_1 = require("./commons");
const crypto_1 = require("./crypto");
const data_1 = require("./data");
const fs_1 = require("./fs");
const tscharset_1 = require("./tscharset");
const tserrors_1 = require("./tserrors");
const types_1 = require("./types");
const customInspectSymbol = Symbol.for('nodejs.util.inspect.custom');
;
class TSData {
    constructor(source, opts = {}) {
        this._allocFn = (0, commons_1.$ok)(opts.allocMethod) ? opts.allocMethod : (opts.fillWithZeros ? Buffer.alloc : Buffer.allocUnsafe);
        if (!(0, commons_1.$ok)(source)) {
            source = 0;
        }
        if (source instanceof TSData) {
            this._len = source._len;
            this._buf = this._allocFn(source.capacity);
            if (this._len > 0) {
                source._buf.copy(this._buf, 0, 0, this._len);
            }
        }
        else if (source instanceof Buffer) {
            this._len = source.length;
            if (opts.dontCopySourceBuffer) {
                this._buf = source;
            }
            else {
                this._buf = this._allocFn((0, commons_1.$capacityForCount)(this._len));
                source.copy(this._buf);
            }
        }
        else if (source instanceof ArrayBuffer) {
            this._buf = (0, data_1.$bufferFromArrayBuffer)(source);
            this._len = this._buf.length;
        }
        else if ((0, commons_1.$isarray)(source) || source instanceof Uint8Array) {
            const slen = source.length;
            this._len = 0;
            this._buf = this._allocFn((0, commons_1.$capacityForCount)(slen));
            this.appendBytes(source, 0, slen);
        }
        else if ((0, commons_1.$isnumber)(source)) {
            const capacity = (0, commons_1.$capacityForCount)(source);
            this._len = 0;
            this._buf = this._allocFn(capacity);
        }
        else {
            throw new tserrors_1.TSError('TSData.constructor() : Bad parameters', { arguments: Array.from(arguments) });
        }
    }
    static fromFile(src) {
        tserrors_1.TSError.assertNotInBrowser('TSData.fromFile');
        const b = (0, fs_1.$readBuffer)(src);
        return (0, commons_1.$ok)(b) ? new TSData(b, { dontCopySourceBuffer: true }) : null;
    }
    static fromString(source, encoding, sourceStart, sourceEnd) {
        const [, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        return start < end ? (0, tscharset_1.$charset)(encoding, tscharset_1.TSCharset.binaryCharset()).dataFromString(source, start, end) : new TSData();
    }
    leafInspect() { return '<' + (0, data_1.$dataAspect)(this.mutableBuffer, { name: this.constructor.name, prefix: '', suffix: '', separator: '', showLength: false, transformFn: (n) => n.toHex2() }) + '>'; }
    [customInspectSymbol](depth, inspectOptions, inspect) {
        return this.leafInspect();
    }
    [Symbol.iterator]() {
        let pos = 0;
        return { next: () => {
                return pos < this._len ? { done: false, value: this._buf[pos++] } : { done: true, value: NaN };
            } };
    }
    splice(targetStart, deleteCount, source, sourceStart, sourceEnd, paddingByte) {
        const datasource = source instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(source) : source;
        return this._splice(targetStart, deleteCount, datasource, sourceStart, sourceEnd, paddingByte);
    }
    appendByte(source) {
        this._willGrow(1);
        this._buf[this._len++] = source & 0xff;
        return this;
    }
    appendBytes(source, sourceStart, sourceEnd) { return (0, commons_1.$ok)(source) ? this._splice(this._len, 0, source, sourceStart, sourceEnd) : this; }
    appendData(source, sourceStart, sourceEnd) {
        if ((0, commons_1.$ok)(source)) {
            const datasource = source instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(source) : source;
            return this._splice(this._len, 0, datasource, sourceStart, sourceEnd);
        }
        return this;
    }
    appendString(source, encoding, sourceStart, sourceEnd) {
        const [, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        if (start < end) {
            const b = (0, tscharset_1.$charset)(encoding, tscharset_1.TSCharset.binaryCharset()).stringToBytes(source, start, end);
            this.appendBytes(b);
        }
        return this;
    }
    replaceBytes(source, targetStart, sourceStart, sourceEnd) {
        const [, start, end, len] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        return this._splice((0, commons_1.$tounsigned)(targetStart), len, source, start, end);
    }
    replaceData(source, targetStart, sourceStart, sourceEnd) {
        const datasource = source instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(source) : source;
        const [, start, end, len] = (0, commons_1.$lse)(datasource, sourceStart, sourceEnd);
        return this._splice((0, commons_1.$tounsigned)(targetStart), len, datasource, start, end);
    }
    replaceString(source, targetStart, encoding, sourceStart, sourceEnd) {
        const datasource = (0, commons_1.$ok)(source) ? (0, tscharset_1.$charset)(encoding, tscharset_1.TSCharset.binaryCharset()).bufferFromString(source, sourceStart, sourceEnd) : null;
        const [, start, end, len] = (0, commons_1.$lse)(datasource);
        return this._splice((0, commons_1.$tounsigned)(targetStart), len, datasource, start, end);
    }
    writeBigInt64BE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeBigInt64BE); }
    writeBigInt64LE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeBigInt64LE); }
    writeBigUInt64BE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeBigUInt64BE); }
    writeBigUInt64LE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeBigUInt64LE); }
    writeUInt8(value, offset) { return this._write(value, offset, 1, Buffer.prototype.writeUInt8); }
    writeUInt16LE(value, offset) { return this._write(value, offset, 2, Buffer.prototype.writeUInt16LE); }
    writeUInt16BE(value, offset) { return this._write(value, offset, 2, Buffer.prototype.writeUInt16BE); }
    writeUInt32LE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeUInt32LE); }
    writeUInt32BE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeUInt32BE); }
    writeInt8(value, offset) { return this._write(value, offset, 1, Buffer.prototype.writeInt8); }
    writeInt16LE(value, offset) { return this._write(value, offset, 2, Buffer.prototype.writeInt16LE); }
    writeInt16BE(value, offset) { return this._write(value, offset, 2, Buffer.prototype.writeInt16BE); }
    writeInt32LE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeInt32LE); }
    writeInt32BE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeInt32BE); }
    writeFloatLE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeFloatLE); }
    writeFloatBE(value, offset) { return this._write(value, offset, 4, Buffer.prototype.writeFloatBE); }
    writeDoubleLE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeDoubleLE); }
    writeDoubleBE(value, offset) { return this._write(value, offset, 8, Buffer.prototype.writeDoubleBE); }
    removeTraillingNewLines() { while (this._len > 0 && this._buf[this._len - 1].isNewLine()) {
        this._len--;
    } ; return this; }
    removeTraillingSpaces() { while (this._len > 0 && this._buf[this._len - 1].isWhiteSpace()) {
        this._len--;
    } ; return this; }
    removeTraillingStrictSpaces() { while (this._len > 0 && this._buf[this._len - 1].isStrictWhiteSpace()) {
        this._len--;
    } ; return this; }
    removeTraillingZeros() { while (this._len > 0 && this._buf[this._len - 1] === 0) {
        this._len--;
    } ; return this; }
    truncateBy(n) {
        if (!(0, commons_1.$isunsigned)(n)) {
            throw new tserrors_1.TSError(`TSDate.truncateBy(${n}) is not valid.`, { data: this, truncateBy: n });
        }
        this.length = n >= this._len ? 0 : this._len - n;
        return this;
    }
    get mutableBuffer() { return this._len === this.capacity ? this._buf : this._buf.subarray(0, this._len); }
    get internalStorage() { return [this._buf, this._len]; }
    set length(n) {
        if (!(0, commons_1.$isunsigned)(n)) {
            throw new tserrors_1.TSError(`TSDate.length = ${n} is not valid.`, { data: this, length: n });
        }
        if (n > this._len) {
            this._willGrow(this._len - n);
            if (this._allocFn !== Buffer.alloc) {
                while (this._len < n) {
                    this._buf[this._len++] = 0;
                }
            }
        }
        else {
            this._len = n;
        }
    }
    get capacity() { return this._buf.length; }
    get length() { return this._len; }
    get byteLength() { return this._len; }
    clone() { return new TSData(this, { allocMethod: this._allocFn }); }
    get buffer() {
        const ret = Buffer.allocUnsafe(this._len);
        if (this._len) {
            this._buf.copy(ret, 0, 0, this._len);
        }
        return ret;
    }
    entries() { return this.mutableBuffer.entries(); }
    keys() { return this.mutableBuffer.keys(); }
    values() { return this.mutableBuffer.values(); }
    includes(value, encoding) {
        if (!(0, commons_1.$ok)(value) || !this._len) {
            return false;
        }
        const data = _dataValue(value, encoding);
        if (!(0, commons_1.$ok)(data)) {
            return false;
        }
        return this._index(data, 0) !== -1;
    }
    startsWith(value, encoding) {
        if (!(0, commons_1.$ok)(value) || !this._len) {
            return false;
        }
        const data = _dataValue(value, encoding);
        if (!(0, commons_1.$ok)(data)) {
            return false;
        }
        return this._index(data, 0) === 0;
    }
    endsWith(value, encoding) {
        if (!(0, commons_1.$ok)(value) || !this._len) {
            return false;
        }
        const data = _dataValue(value, encoding);
        if (!(0, commons_1.$ok)(data)) {
            return false;
        }
        const slen = _searchedLength(data);
        if (slen <= 0) {
            return false;
        }
        const i = this._len - slen;
        if (i < 0) {
            return false;
        }
        return this._index(data, i) === i;
    }
    indexOf(value, byteOffset, encoding) {
        if (!(0, commons_1.$ok)(value) || !this._len) {
            return -1;
        }
        const data = _dataValue(value, encoding);
        if (!(0, commons_1.$ok)(data)) {
            return -1;
        }
        return this._index(data, (0, commons_1.$tounsigned)(byteOffset));
    }
    lastIndexOf(value, byteOffset, encoding) {
        if (!(0, commons_1.$ok)(value) || !this._len) {
            return -1;
        }
        const data = _dataValue(value, encoding);
        if (!(0, commons_1.$ok)(data)) {
            return -1;
        }
        return this._rindex(data, (0, commons_1.$tounsigned)(byteOffset, (this._len - 1)));
    }
    slice(sourceStart, sourceEnd) {
        const [, start, end, len] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        if (len) {
            const ret = new TSData(len, { allocMethod: this._allocFn });
            ret._len = len;
            if (len > 0) {
                this._buf.copy(ret._buf, 0, start, end);
            }
            return ret;
        }
        return new TSData(0);
    }
    copy(targetBuffer, targetStart, sourceStart, sourceEnd) {
        const [, start, end, len] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        if (len) {
            targetStart = (0, commons_1.$tounsigned)(targetStart);
            if (targetBuffer instanceof Uint8Array || targetBuffer instanceof Buffer) {
                return this._buf.copy(targetBuffer, targetStart, start, end);
            }
            targetBuffer.replaceBytes(this._buf, targetStart, start, end);
            return len;
        }
        return 0;
    }
    writeToFile(path, opts) {
        tserrors_1.TSError.assertNotInBrowser('TSData.writeToFile');
        return (0, fs_1.$writeBuffer)(path, this, opts);
    }
    fullWriteToFile(path, opts) {
        tserrors_1.TSError.assertNotInBrowser('TSData.fullWriteToFile');
        return (0, fs_1.$fullWriteBuffer)(path, this, opts);
    }
    equals(otherBuffer) { return this.isEqual(otherBuffer); }
    readBigUInt64BE(offset) { return this._read(offset, 8, Buffer.prototype.readBigUInt64BE); }
    readBigUInt64LE(offset) { return this._read(offset, 8, Buffer.prototype.readBigUInt64LE); }
    readBigInt64BE(offset) { return this._read(offset, 8, Buffer.prototype.readBigInt64BE); }
    readBigInt64LE(offset) { return this._read(offset, 8, Buffer.prototype.readBigInt64LE); }
    readUInt8(offset) { return this._read(offset, 1, Buffer.prototype.readUInt8); }
    readUInt16LE(offset) { return this._read(offset, 2, Buffer.prototype.readUInt16LE); }
    readUInt16BE(offset) { return this._read(offset, 2, Buffer.prototype.readUInt16BE); }
    readUInt32LE(offset) { return this._read(offset, 4, Buffer.prototype.readUInt32LE); }
    readUInt32BE(offset) { return this._read(offset, 4, Buffer.prototype.readUInt32BE); }
    readInt8(offset) { return this._read(offset, 1, Buffer.prototype.readInt8); }
    readInt16LE(offset) { return this._read(offset, 2, Buffer.prototype.readInt16LE); }
    readInt16BE(offset) { return this._read(offset, 2, Buffer.prototype.readInt16BE); }
    readInt32LE(offset) { return this._read(offset, 4, Buffer.prototype.readInt32LE); }
    readInt32BE(offset) { return this._read(offset, 4, Buffer.prototype.readInt32BE); }
    readFloatLE(offset) { return this._read(offset, 4, Buffer.prototype.readFloatLE); }
    readFloatBE(offset) { return this._read(offset, 4, Buffer.prototype.readFloatBE); }
    readDoubleLE(offset) { return this._read(offset, 8, Buffer.prototype.readDoubleLE); }
    readDoubleBE(offset) { return this._read(offset, 8, Buffer.prototype.readDoubleBE); }
    toBase64() { return (0, data_1.$encodeBase64)(this.mutableBuffer); }
    hash(method) { return (0, crypto_1.$hash)(this, method); }
    crc32() { return (0, crypto_1.$crc32)(this); }
    toString(encoding, sourceStart, sourceEnd) {
        return (0, tscharset_1.$charset)(encoding, tscharset_1.TSCharset.binaryCharset()).stringFromData(this, sourceStart, sourceEnd);
    }
    toJSON() { return this.mutableBuffer.toJSON(); }
    toBytes(sourceStart, sourceEnd) {
        return this.toArray(sourceStart, sourceEnd);
    }
    toArray(sourceStart, sourceEnd) {
        const [, start, end] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        return (0, data_1.$bytesFromBytes)(this._buf, start, end);
    }
    toArrayBuffer(sourceStart, sourceEnd) {
        const [, start, end] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        return (0, data_1.$arrayBufferFromBytes)(this._buf, start, end);
    }
    toUint8Array(sourceStart, sourceEnd) {
        const [, start, end] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        return (0, data_1.$uint8ArrayFromBytes)(this._buf, start, end);
    }
    toBuffer(sourceStart, sourceEnd) {
        const [, start, end, len] = (0, commons_1.$lse)(this, sourceStart, sourceEnd);
        const ret = Buffer.allocUnsafe(len);
        if (len > 0) {
            this._buf.copy(ret, 0, start, end);
        }
        return ret;
    }
    compare(other) {
        if (this === other) {
            return types_1.Same;
        }
        else if (other instanceof TSData) {
            return Buffer.compare(this.mutableBuffer, other.mutableBuffer);
        }
        else if (other instanceof Uint8Array) {
            return Buffer.compare(this.mutableBuffer, other);
        }
        else if (other instanceof ArrayBuffer) {
            return this.compare((0, data_1.$bufferFromArrayBuffer)(other));
        }
        return undefined;
    }
    isEqual(other) {
        if (this === other) {
            return true;
        }
        if (other instanceof TSData) {
            if (other._len !== this._len) {
                return false;
            }
            return Buffer.compare(this.mutableBuffer, other.mutableBuffer) === 0;
        }
        else if (other instanceof Uint8Array || other instanceof Buffer) {
            if (other.length !== this._len) {
                return false;
            }
            return Buffer.compare(this.mutableBuffer, other) === 0;
        }
        else if (other instanceof ArrayBuffer) {
            return this.isEqual((0, data_1.$bufferFromArrayBuffer)(other));
        }
        return false;
    }
    _splice(targetStart, deleteCount, source, sourceStart, sourceEnd, paddingByte) {
        const [, start, end, len] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        const padding = Math.min((0, commons_1.$tounsigned)(paddingByte), types_1.UINT8_MAX);
        targetStart = (0, commons_1.$tounsigned)(targetStart);
        deleteCount = (0, commons_1.$tounsigned)(deleteCount);
        if (targetStart + deleteCount >= this._len) {
            if (targetStart + len > this._len) {
                this._willGrow(targetStart + len - this._len);
            }
            for (let i = this._len; i < targetStart; i++) {
                this._buf[i] = padding;
            }
            this._insideCopy(source, start, end, targetStart);
            this._len = targetStart + len;
        }
        else if (deleteCount >= len) {
            this._insideCopy(source, start, end, targetStart);
            if (deleteCount > len) {
                let dstart = targetStart + len;
                for (let i = targetStart + deleteCount; i < this._len; i++) {
                    this._buf[dstart++] = this._buf[i];
                }
                this._len -= deleteCount - len;
            }
        }
        else {
            const glen = len - deleteCount;
            this._willGrow(glen);
            this._insideCopy(source, start, start + deleteCount, targetStart);
            const tend = targetStart + len;
            for (let i = this._len + glen - 1; i >= tend; i--) {
                this._buf[i] = this._buf[i - glen];
            }
            this._insideCopy(source, start + deleteCount, end, targetStart + deleteCount);
            this._len += glen;
        }
        return this;
    }
    _index(value, byteOffset) {
        if (!this._len) {
            return -1;
        }
        const slen = _searchedLength(value);
        if (slen <= 0 || byteOffset + slen > this._len) {
            return -1;
        }
        return this.mutableBuffer.indexOf(value, byteOffset);
    }
    _rindex(value, byteOffset) {
        if (!this._len) {
            return -1;
        }
        const slen = _searchedLength(value);
        if (byteOffset >= this._len) {
            byteOffset = (this._len - 1);
        }
        if (slen <= 0 || byteOffset - slen + 1 < 0) {
            return -1;
        }
        return this.mutableBuffer.lastIndexOf(value, byteOffset);
    }
    _insideCopy(source, start, end, targetStart) {
        if (start < end) {
            if (source instanceof Buffer || source instanceof TSData) {
                source.copy(this._buf, targetStart, start, end);
            }
            else {
                for (let i = start; i < end; i++) {
                    this._buf[targetStart + i] = source[i] & 0xff;
                }
                ;
            }
        }
    }
    _willGrow(n) {
        if (n > 0 && this._len + n > this.capacity) {
            const newCapacity = (0, commons_1.$capacityForCount)((this._len + n));
            let newBuffer = this._allocFn(newCapacity);
            if (this._len > 0) {
                this._buf.copy(newBuffer, 0, 0, this._len);
            }
            this._buf = newBuffer;
        }
    }
    _read(offset = 0, size, bufferReadFn) {
        if (!(0, commons_1.$isunsigned)(offset) || offset + size > this._len) {
            throw new tserrors_1.TSError(`TSData._read(${offset}) out of bound [0,${this._len}]`, { data: this, offset: offset, size: size, readFunction: bufferReadFn });
        }
        return bufferReadFn.call(this._buf, offset);
    }
    _write(value, offset = 0, size, bufferWriteFn) {
        if (!(0, commons_1.$isunsigned)(offset)) {
            throw new tserrors_1.TSError(`TSData._write(${value}, ${offset}) out of bound [0,${this._len}]`, { data: this, offset: offset, size: size, readFunction: bufferWriteFn });
        }
        if (offset + size > this._len) {
            this._willGrow(offset + size - this._len);
        }
        for (let i = this._len; i < offset; i++) {
            this._buf[i] = 0;
        }
        bufferWriteFn.call(this._buf, value, offset);
        return this;
    }
}
exports.TSData = TSData;
function _dataValue(value, encoding) {
    if (value instanceof Buffer || value instanceof Uint8Array) {
        return value;
    }
    if ((0, commons_1.$isstring)(value)) {
        return (0, tscharset_1.$charset)(encoding, tscharset_1.TSCharset.binaryCharset()).uint8ArrayFromString(value);
    }
    if (value instanceof ArrayBuffer) {
        return (0, data_1.$bufferFromArrayBuffer)(value);
    }
    if (value instanceof TSData) {
        return value.mutableBuffer;
    }
    if ((0, commons_1.$isunsigned)(value, 0xff)) {
        return value;
    }
    if ((0, commons_1.$isarray)(value)) {
        return (0, data_1.$uint8ArrayFromBytes)(value);
    }
    return null;
}
function _searchedLength(value) {
    if ((0, commons_1.$isnumber)(value)) {
        return (0, commons_1.$isunsigned)(value, types_1.UINT8_MAX) ? 1 : -1;
    }
    return (0, commons_1.$ok)(value) ? value.length : -1;
}
//# sourceMappingURL=tsdata.js.map